/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2013-2014 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of the GNU General
 * Public License Version 2 only ("GPL"). You may not use this file except
 * in compliance with the License.  You can obtain a copy of the License at
 * https://avatar.java.net/license.html or legal/LICENSE.txt.  See the
 * License for the specific language governing permissions and limitations
 * under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 */


/**
 * This module provides common utility functionality.
 *
 * @module org/glassfish/avatar/common
 */

'use strict';

exports.inherits = function (constructor, superConstructor) {
    constructor.super_ = superConstructor;
    constructor.prototype = Object.create(superConstructor.prototype, {
        constructor: {
            value: constructor,
            enumerable: false,
            writable: true,
            configurable: true
        }
    });
    return constructor;
};

var events = require('events');

var inheritEventEmitter = function (constructor) {
    return exports.inherits(constructor, events.EventEmitter);
};

var appDesc = serviceContainer.getApplicationDescriptor();
var threadIndex = appDesc.threadIndex();

// Enables a JS type to be an EventEmitter that receives events
// from Java EventSource instances. A null source is allowed for
// instances that cannot generate events.

var EventSource = inheritEventEmitter(function (source, handler) {
    var self = this;
    events.EventEmitter.call(self);
    if (source) {
        self._emit = function (event, args) {
            if (!args) {
                self.emit(event);
            } else {
                switch (args.length) {
                    case 0:
                        self.emit(event);
                        break;
                    case 1:
                        self.emit(event, args[0]);
                        break;
                    case 2:
                        self.emit(event, args[0], args[1]);
                        break;
                    case 3:
                        self.emit(event, args[0], args[1], args[2]);
                        break;
                    default:
                        var len = args.length;
                        var newArgs = new Array(len + 1);
                        newArgs[0] = event;
                        for (var i = 0; i < len; i++) {
                            newArgs[i + 1] = args[i];
                        }
                        self.emit.apply(self, newArgs);
                }
            }
        };

        self._relay = source.connect(threadIndex, handler ? handler : self._emit);

        self.on('newListener', function (type) {
            self._relay.listenerAdded(type);
        });

        self.on('removeListener', function (type) {
            self._relay.listenerRemoved(type);
        });
    }
});

exports.EventSource = EventSource;

exports.inheritEventSource = function (constructor) {
    return exports.inherits(constructor, EventSource);
};

var threads = function () {
    if (!threadModule) {
        threadModule = require('org/glassfish/avatar/threads')
    }
    return threadModule;
};

var options, threadModule;   // lazy

/**
 * Provides application information and access to useful instances.
 * @constant
 * @property {string} name The application name.
 * @property {number} threadIndex The index of the current thread.
 * @property {number} threadCount The current number of threads in this application.
 * @property {Date} startTime When the application started.
 * @property {long} uptime The number of milliseconds this application has been running.
 * @property {MessageBus} bus The <a href="threads-MessageBus.html">message bus</a> instance.
 * @property {Cluster} cluster The <a href="threads-Cluster.html">cluster</a> instance.
 * @property {object} options The avatar.properties instance (e.g. options.x returns 'x' value,
 * options['some.key'] returns 'some.key' value).
 * @property {SharedState} state The application scoped <a href="threads-SharedState.html">shared state</a> instance.
 * @property {SessionState} sessions The <a href="threads-SessionState.html">accessor</a> for session scoped
 * <a href="threads-SharedState.html">shared state</a> instances.
 */
exports.application = Object.freeze({
        get name() {
            return appDesc.name();
        },
        get threadIndex() {
            return appDesc.threadIndex();
        },
        get threadCount() {
            return appDesc.threadCount();
        },
        get startTime() {
            return new Date(appDesc.startTime());
        },
        get uptime() {
            return appDesc.uptime();
        },
        get bus() {
            return (threads()).messageBus();
        },
        get state() {
            return (threads()).applicationState();
        },
        get sessions() {
            return (threads()).sessionState();
        },
        get cluster() {
            return (threads()).cluster();
        },
        get options() {
            if (!options) {
                options = servicePool.getOptions();
            }
            return {
                __noSuchProperty__: function (name) {
                    return options.get(name);
                }
            }
        }
    }
);

var blockingLogger = Java.type('org.glassfish.avatar.util.Log');
var logger = logger ? logger : blockingLogger;
var jsToString = Java.type('jdk.nashorn.internal.runtime.JSType').toString;

/**
 * Log a message.
 * @param {string} msg The message.
 */
exports.log = function (msg) {
    logger.log(jsToString(msg));
};

/**
 * Returns true if FINE logging is enabled.
 * @returns {boolean}
 */
exports.willLogFine = function () {
    return blockingLogger.willLogFine();
};

/**
 * Log a message if logger is at level FINE or below.
 * @param {string} msg The message.
 */
exports.logFine = function (msg) {
    if (blockingLogger.willLogFine()) {
        logger.log(jsToString(msg));
    }
};

/**
 * Log a warning message.
 * @param {string} msg The message.
 * @param {object} exception The exception, if any.
 */
exports.logWarning = function (msg, exception) {
    var jsMsg = jsToString(msg);
    if (logger === blockingLogger) {
        blockingLogger.logWarning(jsMsg, exception);
    } else if (exception) {
        logger.log("WARNING: " + jsMsg + " (caught " + jsToString(exception) + ")");
    } else {
        logger.log("WARNING: " + jsMsg);
    }
};

/**
 * Returns the specified module, if available, or a default if not.
 * @param {string} id The module id.
 * @param {string} defaultMessage An optional message to log if the module is not available.
 * @param {object} defaultModule The module to return if the specified module is not available.
 * @returns {object} The specified or default module.
 */
exports.optionalModule = function (id, defaultMessage, defaultModule) {
    try {
        return require(id);
    } catch (e) {
        exports.log(defaultMessage + '(' + e + ')');
        return defaultModule;
    }
};

/**
 * The promise type.
 * @see {@link https://github.com/cujojs/when when.js}
 */
var Promise = require('when');
exports.Promise = Promise;

/**
 * Wraps an optional callback in a promise.
 * @param {function} callback An optional callback.
 * @param {Promise} promise The promise;
 * @returns Either the original promise if no callback, or a new one that wraps the callback.
 */
exports.wrapCallback = function (callback, promise) {
    if (callback) {
        promise.then(
            function (args) {
                //Add an empty error parameter at the beginning of argument list
                callback.apply(null, [null].concat([args]));
            }, callback);
    } else {
        return promise;
    }
};

/**
 * Calls an async node style function that takes a callback, and returns a promise for the result.
 * Pass the function as the first argument, remaining args are passed to the function.
 */
exports.nodeCall = function () {
    var deferred = Promise.defer();
    var args = Array.prototype.slice.call(arguments);
    var fn = args.shift();
    args.push(function (error, result) {
        if (error) {
            error.printStackTrace();
            deferred.reject(error);
        } else {
            deferred.resolve(result);
        }
    });
    fn.apply(null, args);
    return deferred.promise;
};

/**
 * Wrapper function to catch errors both synchronously and asynchronously using promise.
 * @param {function} fn Function to call, may return a promise.
 * @param {function} fail Failure handler to call in case on an exception from f or from the promise
 * returned by f.
 */
exports.tryAndCatch = function (fn, fail) {
    try {
        var retval = fn();
        if (retval && Promise.isPromise(retval)) {
            retval.then(function (value) { /*This is unnecessary but for some reason needed by when.js*/
            }, fail);
        }
    } catch (error) {
        fail(error);
    }
};
